import time, sys, os
import shlex
import numpy as np
from subprocess import check_output, PIPE, Popen

K = 200
least_reads = 2
POA = ''

'''
class: Gene
'''
class Gene():
    def __init__(self, id, name, start, end, chr, flag):
        self.id = id
        self.name = name
        self.start = start
        self.end = end
        self.chr = chr
        self.flag = flag
        pass
'''
class: Fusion
'''
class Fusion():
    def __init__(self, gene1, gene2, num, bp1, bp2, code, reads, only_id):
        self.gene1 = gene1
        self.gene2 = gene2
        self.num = num
        self.bp1 = bp1
        self.bp2 = bp2
        self.code = code
        self.reads = reads
        self.read_id = self.gene1.id + '_' + self.gene2.id + '_' + code
        self.only_id = only_id
        self.flag = 2
        pass
    def add_result(self, result):
        self.result = result
        self.flag = 3
'''
class: Result
'''
class Result():
    def __init__(self, only_id, sequence, samline):
        self.only_id = only_id
        self.transcript = sequence
        self.samlines = [samline,]
        self.num = samline.num
        self.flag = 3
        self.reads = ''
        pass
    def add_samline(self, samline):
        self.samlines.append(samline)
        pass
    def add_fusion(self, fusion):
        self.fusion = fusion
        self.reads = fusion.reads
        pass
    def add_bp(self, bp1s, bp2s):
        self.bp1 = int(np.mean(bp1s))
        self.bp2 = int(np.mean(bp2s))
        pass
    
'''
class: samline
'''
class Samline():
    def __init__(self, read, chr, flag, start, cigar, num):
        self.read = read
        self.chr = chr
        self.start = start
        self.cigar = cigar
        self.flag = flag
        self.num = num
        self.compute_end()
        self.compute_position()
        pass
    def compute_end(self):
        num = ''
        length = 0
        for i in self.cigar:
            if i.isdigit():
                num = num + i
            elif i == 'I' or i == 'S' or i == 'H':
                num = ''
            else:
                length = length + int(num)
                num = ''
        self.length = length
        self.end = self.start + length - 1
        pass
    def compute_position(self):
        num = ''
        length = 0
        for i in self.cigar:
            if i.isdigit():
                num = num + i
            elif i == 'D'  or i == 'N':
                num = ''
            else:
                length = length + int(num)
                num = ''
        self.ownlength = length
        if 'S' in self.cigar or 'H' in self.cigar:
            if 'S' in self.cigar:
                key = 'S'
            elif 'H' in self.cigar:
                key = 'H'
            num = len(self.cigar.split(key))
            if num == 2:
                if self.cigar.split(key)[1] == '':
                    self.position1 = 0
                    num = ''
                    for i in self.cigar.split(key)[0][::-1]:
                        if i.isdigit():
                            num += i
                        else:
                            break
                    num = int(num[::-1])
                    self.position2 = length - num
                else:
                    self.position1 = int(self.cigar.split(key)[0])
                    self.position2 = length
            else:
                self.position1 = int(self.cigar.split(key)[0])
                num = ''
                for i in self.cigar.split(key)[1][::-1]:
                    if i.isdigit():
                        num += i
                    else:
                        break
                num = int(num[::-1])
                self.position2 = length - num
        else:
            self.position1 = 0
            self.position2 = length
        if self.flag == 16 or self.flag == 2064 or self.flag == 272:
            position1 = self.ownlength - self.position1 + 1
            position2 = self.ownlength - self.position2 + 1
            self.position1 = position2
            self.position2 = position1
        pass

'''
fuction: Antisense
input: string
output: antistring
'''
def Antisense(string):
    antistring = ''
    for i in string:
        if i == 'A':
            antistring += 'T'
        elif i == 'T':
            antistring += 'A'
        elif i == 'C':
            antistring += 'G'
        elif i == 'G':
            antistring += 'C'
    antistring = antistring[::-1]
    return antistring
'''
fuction: subprocess_popen
'''
def subprocess_popen(args, stdin=None, stdout=PIPE, stderr=sys.stderr, bufsize=8388608):
    return Popen(args, stdin=stdin, stdout=stdout, stderr=stderr, bufsize=bufsize, universal_newlines=True)
'''
fuction: read_result
'''
def read_result(result_file):
    fusions = {}
    for line in result_file.split('\n')[:-1]:
        n1 = line.split('\t')[0]
        n2 = line.split('\t')[1]
        id1 = line.split('\t')[2]
        id2 = line.split('\t')[3]
        chr1 = line.split('\t')[4].split(':')[0]
        flag1 = line.split('\t')[4].split(':')[1]
        chr2 = line.split('\t')[5].split(':')[0]
        flag2 = line.split('\t')[5].split(':')[1]
        start1 = int(line.split('\t')[6].split(':')[0])
        end1 = int(line.split('\t')[6].split(':')[1])
        start2 = int(line.split('\t')[7].split(':')[0])
        end2 = int(line.split('\t')[7].split(':')[1])
        bp1 = int(line.split('\t')[8])
        bp2 = int(line.split('\t')[9])
        num = int(line.split('\t')[10])
        code = line.split('\t')[11]
        reads = line.split('\t')[12].split('\n')[0]
        only_id = id1 + '_' + id2 + '_' + code + '_' + str(num)
        gene1 = Gene(id1, n1, start1, end1, chr1, flag1)
        gene2 = Gene(id2, n2, start2, end2, chr2, flag2)
        fusion = Fusion(gene1, gene2, num, bp1, bp2, code, reads, only_id)
        if id1 + '_' + id2 in fusions:
            fusions[id1 + '_' + id2].append(fusion)
        else:
            fusions[id1 + '_' + id2] = [fusion,]
    return fusions
def read_result0():
    file = open('./middlefile/step1.genefusion.csv')
    line = file.readline()
    linelist = file.readlines()
    result = ''
    for line in linelist:
        result += line
    return result

'''
fuction: alignment
'''
def alignment(POA_file, index_dir, middlefile, thread, if_output):
    results_of_each_fusion = {} #{key = key_id : value = [result,...]}
    ref_index = index_dir + 'ref_index.mmi'
    minimap2_view_process = subprocess_popen(shlex.split('minimap2 -ax splice -uf -k14  -t {} {} {}'.format(str(thread), ref_index, POA_file)))
    with open(middlefile + 'step2.RNA_alignment.sam', 'w') as samfile:
        read0 = ''
        for line in minimap2_view_process.stdout:
            if line != '' and line != None:
                if if_output == True:
                    samfile.write(line)
                if '@' not in line:
                    flag = int(line.split('\t')[1])
                    if flag != 4:
                        read = line.split('\t')[0]
                        if read != read0:
                            read0 = read
                            id1 = read.split('_')[0]
                            id2 = read.split('_')[1]
                            num = int(read.split('_')[3])
                            read_id = id1 + '_' + id2
                            chr = line.split('\t')[2]
                            start = int(line.split('\t')[3])
                            cigar = line.split('\t')[5]
                            if flag == 0:
                                sequence = line.split('\t')[9]
                            elif flag == 16:
                                sequence = Antisense(line.split('\t')[9])
                            samline = Samline(read, chr, flag, start, cigar, num)
                            result = Result(read, sequence, samline)
                            if read_id in results_of_each_fusion:
                                results_of_each_fusion[read_id].append(result)
                            else:
                                results_of_each_fusion[read_id] = [result,]
                        else:
                            id1 = read.split('_')[0]
                            id2 = read.split('_')[1]
                            num = int(read.split('_')[3])
                            read_id = id1 + '_' + id2
                            chr = line.split('\t')[2]
                            start = int(line.split('\t')[3])
                            cigar = line.split('\t')[5]
                            samline = Samline(read, chr, flag, start, cigar, num)
                            result.add_samline(samline)
    return results_of_each_fusion

'''
fuction: outputline
'''
def outputline(fusion):
    l1 = fusion.gene1.name
    l2 = fusion.gene2.name
    l3 = fusion.gene1.id
    l4 = fusion.gene2.id
    l5 = fusion.gene1.chr + ':' + str(fusion.bp1) + '; ' + fusion.gene2.chr + ':' + str(fusion.bp2)
    l6 = fusion.gene1.chr + fusion.gene1.flag + ': ' + str(fusion.gene1.start) + ', ' + str(fusion.gene1.end)
    l7 = fusion.gene2.chr + fusion.gene2.flag + ': ' + str(fusion.gene2.start) + ', ' + str(fusion.gene2.end)
    l10 = fusion.reads
    if fusion.flag == 2:
        flag = 3
        l8 = str(fusion.num)
        l9 = 'flag = 3: Potential Fusion'
    if fusion.flag == 3:
        if fusion.result.flag == 3:
            flag = 3
            l8 = str(fusion.num)
            l9 = 'flag = 3: Potential Fusion'
        elif fusion.result.flag == 4:
            flag = 4
            l8 = str(fusion.num)
            l9 = 'flag = 4: Suspected Fusion'
        elif fusion.result.flag == 5:
            flag = 5
            l5 = fusion.gene1.chr + ':' + str(fusion.result.bp1) + '; ' + fusion.gene2.chr + ':' + str(fusion.result.bp2)
            l8 = str(fusion.num)
            l9 = 'flag = 5: Reliable Fusion'
    line = l1 + '\t' + l2 + '\t' + l3 + '\t' + l4 + '\t' + l5 + '\t' + l6 + '\t' + l7 + '\t' + l8 + '\t' + l9 + '\t' + l10 + '\n'
    return flag, line
        
'''
fuction: filtering_and_output
'''
def filtering_and_output(results_of_each_fusion, fusions_dict, result_out):
    with open(result_out, 'w') as file:
        line = 'gene1 name\tgene2 name\tgene1 id\tgene2 id\tbreak points\tgene1 position\tgene2 position\tsupport num\trank class\tsupperting reads information\n'
        file.write(line)
        for key, fusions in fusions_dict.items():
            if key in results_of_each_fusion:
                gene1 = fusions[0].gene1
                gene2 = fusions[0].gene2
                results = results_of_each_fusion[key]
                for result in results:
                    sam1 = []
                    sam2 = []
                    for samline in result.samlines:
                        if samline.end - gene1.start > 0 and gene1.end - samline.start > 0 and samline.chr == gene1.chr:
                            sam1.append(samline)
                        elif samline.end - gene2.start > 0 and gene2.end - samline.start > 0 and samline.chr == gene2.chr:
                            sam2.append(samline)
                    if sam1 != [] and sam2 != []:
                        result.flag = 4
                        for fusion in fusions:
                            if result.only_id == fusion.only_id:
                                result.add_fusion(fusion)
                                fusion.add_result(result)
                                break
                        for fusion in fusions:
                            bp1s = []
                            bp2s = []
                            for sam in sam1:
                                bp1, f1 = check_breakpoints1(fusion.bp1, sam)
                                if f1 == True:
                                    bp1s.append(bp1)
                            for sam in sam2:
                                bp2, f2 = check_breakpoints2(fusion.bp2, sam)
                                if f2 == True:
                                    bp2s.append(bp2)
                            if bp1s != [] and bp2s != []:
                                result.add_bp(bp1s, bp2s)
                                if result.only_id == fusion.only_id:
                                    result.flag = 5
                                else:
                                    fusion.num += result.num
                                    fusion.reads += result.reads
        for key, fusions in fusions_dict.items():
            for fusion in fusions:
                if (fusion.flag == 2 and fusion.num >= least_reads) or (fusion.flag == 3 and fusion.result.num >= least_reads):
                    flag, line = outputline(fusion)
                    file.write(line)
        '''lines_5 = []
        lines_4 = []
        lines_3 = []
        for key, fusions in fusions_dict.items():
            for fusion in fusions:
                if (fusion.flag == 2 and fusion.num >= least_reads) or (fusion.flag == 3 and fusion.result.num >= least_reads):
                    flag, line = outputline(fusion)
                    if flag == 5:
                        lines_5.append(line)
                    elif flag == 4:
                        lines_4.append(line)
                    elif flag == 3:
                        lines_3.append(line)
        for line in lines_5:            
            file.write(line)
        for line in lines_4:            
            file.write(line)
        for line in lines_3:            
            file.write(line)'''

def check_breakpoints1(bp1, samline):
    flag = False
    bp = 0
    if samline.flag == 0 or samline.flag == 256 or samline.flag == 2048:
        if np.abs(samline.end - bp1) <= K:
            flag = True
            bp = samline.end
    elif samline.flag == 16 or samline.flag == 272 or samline.flag == 2064:
        if np.abs(samline.start - bp1) <= K:
            flag = True
            bp = samline.start
    return (bp, flag)

def check_breakpoints2(bp2, samline):
    flag = False
    bp = 0
    if samline.flag == 0 or samline.flag == 256 or samline.flag == 2048:
        if np.abs(samline.start - bp2) <= K:
            flag = True
            bp = samline.start
    elif samline.flag == 16 or samline.flag == 272 or samline.flag == 2064:
        if np.abs(samline.end - bp2) <= K:
            flag = True
            bp = samline.end
    return (bp, flag)


def main(result_out, POA_file, result_file, index_dir, middlefile, thread, if_output, lreads, bpsbp):
    global least_reads, K
    K = bpsbp
    least_reads = lreads
    fusions = read_result(result_file)
    results_of_each_fusion = alignment(POA_file, index_dir, middlefile, thread, if_output)
    filtering_and_output(results_of_each_fusion, fusions, result_out)

if __name__ == "__main__":
    result_out = './result.csv'
    POA_file = './middlefile/step1.POA_result.fasta'
    result_file = read_result0()
    index_dir = '../GFHunter/index/'
    middlefile = './middlefile/'
    thread = 4
    if_output = True
    main(result_out, POA_file, result_file, index_dir, middlefile, thread, if_output, 2)
    


import time
import os
import argparse
from . import preprocessing
from . import transcriptome_based_detection
from . import crossing_reference_based_refinement
from . import sc_handle


'''
fuction: sub_cmd_index
'''
def sub_cmd_index(args):
    time1 = time.time()
    preprocessing.main(vars(args)['<annotationfile>'], vars(args)['<referencefile>'], vars(args)['<indexdir>'])
    os.system('minimap2 -d ' + vars(args)['<indexdir>'] + 'ref_index.mmi ' + vars(args)['<referencefile>'])
    os.system('minimap2 -d ' + vars(args)['<indexdir>'] + 'trans_index.mmi ' + vars(args)['<indexdir>'] + 'rebuild_transcript.fa')
    time2 = time.time()
    timeuse = time2 - time1
    print('Totally using ' + str(timeuse) + 's')
    print('-' * 50)
'''
fuction: sub_cmd_detect
'''
def sub_cmd_detect(args):
    time1 = time.time()
    if not os.path.exists(args.middlefile):
        os.makedirs(args.middlefile)
    #step1
    print('-' * 50)
    print('TRANSCRIPTOME-BASED DETECTION')
    cnums = (args.min_read_length, args.max_exon_boundary, args.overlap_precent, args.min_clustering_length, args.trans_based_align_type, args.least_support_reads)
    result_file, POA_file, supporting_reads = transcriptome_based_detection.main(vars(args)['<readfile>'], vars(args)['<indexdir>'], args.middlefile, args.threads, args.print_middle_output, cnums, args.countine)
    #step2
    print('-' * 50)
    print('CROSS-REFERENCING-BASED FUSION REFINEMENT')
    print('-' * 50)
    POA_file = args.middlefile + 'step1.POA_result.fasta'
    crossing_reference_based_refinement.main(args.output + '.csv', POA_file, result_file, vars(args)['<indexdir>'], args.middlefile, args.threads, args.print_middle_output, args.least_support_reads, args.max_breakpoint_distance)
    os.system('mv ' + POA_file + ' ' + args.output + '.fusion_transcripts.fasta')
    if args.print_middle_output == False:
        os.system('rm -rf ' + args.middlefile)
    time2 = time.time()
    timeuse = time2 - time1
    print('-' * 50)
    print('Totally used ' + str(timeuse) + 's')
    print('-' * 50)
'''
fuction: sub_cmd_sc
'''
def sub_cmd_sc(args):
    time1 = time.time()
    if not os.path.exists(args.middlefile):
        os.makedirs(args.middlefile)
    #step0
    print('-' * 50)
    print('SINGLE CELL DATA INDEXING')
    sc_handle.read_bam(vars(args)['<indexdir>'], vars(args)['<bamfile>'], args.output)
    #step1
    print('-' * 50)
    print('TRANSCRIPTOME-BASED DETECTION')
    cnums = (args.min_read_length, args.max_exon_boundary, args.overlap_precent, args.min_clustering_length, args.trans_based_align_type, args.least_support_reads)
    result_file, POA_file, supporting_reads = transcriptome_based_detection.main(args.output + '.fasta', vars(args)['<indexdir>'], args.middlefile, args.threads, args.print_middle_output, cnums, args.countine)
    #step2
    print('-' * 50)
    print('CROSS-REFERENCING-BASED FUSION REFINEMENT')
    print('-' * 50)
    POA_file = args.middlefile + 'step1.POA_result.fasta'
    crossing_reference_based_refinement.main(args.output + '.tsv', POA_file, result_file, vars(args)['<indexdir>'], args.middlefile, args.threads, args.print_middle_output, args.least_support_reads, args.max_breakpoint_distance)
    os.system('mv ' + POA_file + ' ' + args.output + '.fusion_transcripts.fasta')
    if args.print_middle_output == False:
        os.system('rm -rf ' + args.middlefile)
    #step3
    print('-' * 50)
    print('MAP FUSION TO CELLS')
    print('-' * 50)
    sc_handle.handle(vars(args)['<indexdir>'], vars(args)['<bamfile>'], args.output, args.least_support_cells)
    time2 = time.time()
    timeuse = time2 - time1
    print('-' * 50)
    print('Totally used ' + str(timeuse) + 's')
    print('-' * 50)

'''
fuction: 
'''
def Parser_set():
    parser = argparse.ArgumentParser()
    subparser = parser.add_subparsers()

    index_parser = subparser.add_parser('index', help = 'Create index of the GFHunter')
    index_parser.add_argument('<annotationfile>', type= str, help= 'Gene annotation file (gtf)')
    index_parser.add_argument('<referencefile>', type= str, help= 'Reference file (fasta/fa)')
    index_parser.add_argument('<indexdir>', type= str, default='./index/', help= 'Output index directory')
    index_parser.set_defaults(func= sub_cmd_index)

    detect_parser = subparser.add_parser('detect', help = 'Dectect gene fusions')
    detect_parser.add_argument('<readfile>', type= str, help= 'Read file (fasta/fastq)')
    detect_parser.add_argument('<indexdir>', type= str, default='./index/', help= 'Index directory')
    detect_parser.add_argument('-o', '--output', type= str, default= './result', help= 'the name of fusion result (default = "./result")', metavar= 'str')
    detect_parser.add_argument('-m', '--middlefile', type= str, default= './middlefile/', help= 'temporary folder of middle files (default = "./middlefile/")', metavar= 'dir')
    detect_parser.add_argument('-M', '--print_middle_output', action= 'store_true', help= 'retain the middle files in detection')
    detect_parser.add_argument('-t', '--threads', type= int, default= 4, help= 'threads GFHunter used (default = 4)', metavar= 'int')
    detect_parser.add_argument('-T', '--trans_based_align_type', type= str, default= 'ont', help= 'setting the transcriptome-based alignmnet type of minimap2: pb/hifi/ont/iclr - CLR/HiFi/Nanopore/ICLR vs reference mapping (default = ont)', metavar= 'type')
    detect_parser.add_argument('-n', '--min_read_length', type= int, default= 50, help= 'minimum length of reads (bp) considered (default = 50)', metavar= 'int')
    detect_parser.add_argument('-e', '--max_exon_boundary', type= int, default= 30, help= 'maximum length between breakpoint and exon boundary (default = 30)', metavar= 'int')
    detect_parser.add_argument('-p', '--overlap_precent', type= float, default= 0.5, help= 'precent of overlap between reads and transcripts (default = 0.5)', metavar= 'float')
    detect_parser.add_argument('-c', '--min_clustering_length', type= int, default= 200, help= 'minimum distance between two cluster (default = 200)', metavar= 'int')
    detect_parser.add_argument('-l', '--least_support_reads', type= int, default= 2, help= 'least reads number to support gene fusions (default = 2)', metavar= 'int')
    detect_parser.add_argument('-b', '--max_breakpoint_distance', type= int, default= 200, help= 'maximum distance between crossing refinement breakpoints and transcriptome based detection breakpoints (default = 200)', metavar= 'int')
    detect_parser.add_argument('-C', '--countine', action= 'store_true', help= 'countine detection')

    detect_parser.set_defaults(func= sub_cmd_detect)

    sc_parser = subparser.add_parser('sc', help = 'Dectect gene fusions on single cell RNA-seq data')
    sc_parser.add_argument('<bamfile>', type= str, help= 'Read file handled by wf-single-cell (bam)')
    sc_parser.add_argument('<indexdir>', type= str, default='./index/', help= 'Index directory')
    sc_parser.add_argument('-o', '--output', type= str, default= './result', help= 'the name of fusion result (default = "./result")', metavar= 'str')
    sc_parser.add_argument('-m', '--middlefile', type= str, default= './middlefile/', help= 'temporary folder of middle files (default = "./middlefile/")', metavar= 'dir')
    sc_parser.add_argument('-M', '--print_middle_output', action= 'store_true', help= 'retain the middle files in scion')
    sc_parser.add_argument('-t', '--threads', type= int, default= 4, help= 'threads GFHunter used (default = 4)', metavar= 'int')
    sc_parser.add_argument('-T', '--trans_based_align_type', type= str, default= 'ont', help= 'setting the transcriptome-based alignmnet type of minimap2: pb/hifi/ont/iclr - CLR/HiFi/Nanopore/ICLR vs reference mapping (default = ont)', metavar= 'type')
    sc_parser.add_argument('-n', '--min_read_length', type= int, default= 50, help= 'minimum length of reads (bp) considered (default = 50)', metavar= 'int')
    sc_parser.add_argument('-e', '--max_exon_boundary', type= int, default= 30, help= 'maximum length between breakpoint and exon boundary (default = 30)', metavar= 'int')
    sc_parser.add_argument('-p', '--overlap_precent', type= float, default= 0.5, help= 'precent of overlap between reads and transcripts (default = 0.5)', metavar= 'float')
    sc_parser.add_argument('-c', '--min_clustering_length', type= int, default= 200, help= 'minimum distance between two cluster (default = 200)', metavar= 'int')
    sc_parser.add_argument('-l', '--least_support_reads', type= int, default= 2, help= 'least reads number to support gene fusions (default = 2)', metavar= 'int')
    sc_parser.add_argument('-L', '--least_support_cells', type= int, default= 5, help= 'least cells number to support gene fusions (default = 5)', metavar= 'int')
    sc_parser.add_argument('-b', '--max_breakpoint_distance', type= int, default= 200, help= 'maximum distance between crossing refinement breakpoints and transcriptome based scion breakpoints (default = 200)', metavar= 'int')
    sc_parser.add_argument('-C', '--countine', action= 'store_true', help= 'countine detection')

    sc_parser.set_defaults(func= sub_cmd_sc)
    
    args = parser.parse_args()
    if not hasattr(args, 'func'):
        args = parser.parse_args(['-h'])
    args.func(args)
if __name__ == "__main__":
    Parser_set()


import time
import os, sys
import shlex
from subprocess import check_output, PIPE, Popen
from intervaltree import IntervalTree
import numpy as np
from scipy.sparse import csr_matrix
import pyabpoa as pa

min_length = 50
exon_boundary = 30
overlap_present = 0.5
clustering_num = 200
sup_read_num = 2
type_alignment = 'ont'
sam_file = 'middlefile/step1.DNA_like_alignment.only_primary.sam'
output_file_dir = 'middlefile/step1.gene_groups/'
tree_dir = 'middlefile/step1.clustering_tree/'
out_put = 'middlefile/step1.POA_result.fasta'
result_out = 'middlefile/step1.genefusion.csv'
output_sam = 'middlefile/step1.DNA_like_alignment.sam'

if_output = True
thread = 32
min_abpoa = 100
'''
global variable
'''
index = {} #{key = chr : value = IntervalTree_of_transcript_in_the_chr}
reads = [] #[read1, read2, ...]  readn = [read_tuple, ...]
'''
class: Gene
'''
class Gene():
    def __init__(self, id, name, refstart, refend, chr, flag, gene_type):
        self.id = id
        self.name = name
        self.rstart = refstart
        self.rend = refend
        self.chr = chr
        self.flag = flag
        self.type = gene_type
        self.transcripts = []
        pass
    def add_transcript(self, transcript):
        self.transcripts.append(transcript)
        pass
'''
class: Transcript
'''
class Transcript():
    def __init__(self, id, name, refstart, refend, start, end, chr, flag, trans_type, gene):
        self.id = id
        self.name = name
        self.rstart = refstart
        self.rend = refend
        self.start = start
        self.end = end
        self.chr = chr
        self.flag = flag
        self.trans_type = trans_type
        self.gene = gene
        self.exons = []
        pass
    def add_exon(self, exon):
        self.exons.append(exon)
        pass
'''
class: Exon
'''
class Exon():
    def __init__(self, id, refstart, refend, start, end, transcript):
        self.id = id
        self.rstart = refstart
        self.rend = refend
        self.start = start
        self.end = end
        self.transcript = transcript
        self.length = self.end - self.start + 1
        pass
'''
class Samline
'''
class Samline():
    def __init__(self, id, start, cigar, flag, chr, fasta):
        self.id = id
        self.start = start
        self.cigar = cigar
        self.flag = flag
        self.chr = chr
        self.transcripts = []
        self.overlap = 0
        self.transcript = 0
        self.sequence = fasta
        if flag == 0 or flag == 16:
            self.major = True
        else:
            self.major = False
        pass
    def compute_end(self):
        num = ''
        length = 0
        for i in self.cigar:
            if i.isdigit():
                num = num + i
            elif i == 'I' or i == 'S' or i == 'H':
                num = ''
            else:
                length = length + int(num)
                num = ''
        self.length = length
        self.end = self.start + length - 1
        pass
    def compute_position(self):
        num = ''
        length = 0
        for i in self.cigar:
            if i.isdigit():
                num = num + i
            elif i == 'D':
                num = ''
            else:
                length = length + int(num)
                num = ''
        self.ownlength = length
        if 'S' in self.cigar or 'H' in self.cigar:
            if 'S' in self.cigar:
                key = 'S'
            elif 'H' in self.cigar:
                key = 'H'
            num = len(self.cigar.split(key))
            if num == 2:
                if self.cigar.split(key)[1] == '':
                    self.position1 = 0
                    num = ''
                    for i in self.cigar.split(key)[0][::-1]:
                        if i.isdigit():
                            num += i
                        else:
                            break
                    num = int(num[::-1])
                    self.position2 = length - num
                else:
                    self.position1 = int(self.cigar.split(key)[0])
                    self.position2 = length
            else:
                self.position1 = int(self.cigar.split(key)[0])
                num = ''
                for i in self.cigar.split(key)[1][::-1]:
                    if i.isdigit():
                        num += i
                    else:
                        break
                num = int(num[::-1])
                self.position2 = length - num
            if 'S' in self.cigar:
                self.sequence = self.sequence[self.position1:self.position2]
        else:
            self.position1 = 0
            self.position2 = length
        if self.flag == 16 or self.flag == 2064:
            position1 = self.ownlength - self.position1 + 1
            position2 = self.ownlength - self.position2 + 1
            self.position1 = position2
            self.position2 = position1
            self.sequence = Antisense(self.sequence)
    def add_num(self, num):
        self.num = num
    def add_transcript(self, transcript, start, end):
        self.transcripts.append(transcript)
        overlap1 = end - self.start
        overlap2 = self.end - start
        length1 = self.length
        length2 = end - start +1
        real_overlap = min(overlap1, overlap2, length1, length2)
        if self.overlap <= real_overlap:
            self.overlap = real_overlap
            self.transcript = transcript
            self.tstart = start
            self.tend = end
        pass
    def add_read(self, read):
        self.read = read
'''
class: Read
'''
class Read():
    def __init__(self, id, start, cigar, flag, chr, fasta):
        self.id = id
        self.num = 1
        self.samlines = []
        self.sequence = fasta
        if flag == 16:
            self.sequence = Antisense(self.sequence)
        samline = Samline(id, start, cigar, flag, chr, fasta)
        samline.compute_end()
        samline.compute_position()
        samline.add_read(self)
        length = samline.position2 - samline.position1 + 1
        if length > min_length:
            self.samlines.append(samline)
        pass
    def add_samline(self, id, start, cigar, flag, chr, fasta):
        self.num += 1
        samline = Samline(id, start, cigar, flag, chr, fasta)
        samline.compute_end()
        samline.compute_position()
        samline.add_read(self)
        length = samline.position2 - samline.position1 + 1
        if length > min_length:
            self.samlines.append(samline)
        pass
    def num_samline(self):
        samlines = self.samlines
        n = len(samlines)
        for i in range(n - 1):
            for j in range(i + 1, n):
                if samlines[i].position1 > samlines[j].position1:
                    samline = samlines[j]
                    samlines[j] = samlines[i]
                    samlines[i] = samline
        for i in range(n):
            samlines[i].add_num(i)
        pass
    def transcripts_and_genes(self):
        global exon_boundary
        self.transcripts = []
        self.genes = {}
        samlines = []
        self.read_through = []
        poslist = []
        for samline in self.samlines:
            if samline.transcript != 0:                
                if samline.start > samline.transcript.start + exon_boundary or samline.end < samline.transcript.end - exon_boundary: #?????
                    samlines.append(samline)
                else:
                    self.read_through.append(samline)
        if samlines == []:
            return False
        self.samlines = samlines
        for samline in self.samlines:
            if samline.transcript not in self.transcripts:
                self.transcripts.append(samline.transcript)
                if samline.transcript.gene not in self.genes:
                    self.genes[samline.transcript.gene] = [samline.transcript,]
                else:
                    self.genes[samline.transcript.gene].append(samline.transcript)
        return True
    def group_samlines(self, gene1, gene2):
        length1 = 0
        length2 = 0
        for samline in self.samlines:
            if samline.transcript.gene.id == gene1.id:
                if samline.length > length1:
                    length1 = samline.length
                    samline1 = samline
            elif samline.transcript.gene.id == gene2.id:
                if samline.length > length2:
                    length2 = samline.length
                    samline2 = samline
        return (samline1, samline2)
'''
class: Genes_group
'''
class Genes_group():
    def __init__(self, gene1, gene2, reads):
        self.gene1 = gene1
        self.gene2 = gene2
        self.reads = reads
        pass
    def clustering(self):
        global sup_read_num
        self.trees = {}
        self.leaves = {}
        self.flag = False
        n = 0
        for read in self.reads:
            samline1, samline2 = read.group_samlines(self.gene1, self.gene2)
            leaf = Tree_leaf(read, n, samline1, samline2)
            if leaf.flag != 4:
                if leaf.flag in self.leaves:
                    self.leaves[leaf.flag].append(leaf)
                else:
                    self.leaves[leaf.flag] = [leaf,]
                n += 1
        for flag, leaves in self.leaves.items():
            if len(leaves) >= sup_read_num:
                self.flag = True
                self.trees[flag] = Cluster_tree(leaves, self.gene1, self.gene2, flag)
                self.trees[flag].clustering() 
        pass
'''
class: Cluster_tree
'''
class Tree_leaf():
    def __init__(self, read, num, samline1, samline2):
        self.read = read
        self.num = num
        self.samline1 = samline1
        self.samline2 = samline2
        self.bp_position()
        pass
    def add_address(self, address):
        self.address = address
        pass
    def front_bp_pos(self, samline):
        global exon_boundary
        for exon in samline.transcript.exons[:-1]:
            if exon.end - exon_boundary <= samline.end and exon.end + exon_boundary >= samline.end:
                return samline.end - exon.end + exon.rend
        return False
    def behind_bp_pos(self, samline):
        global exon_boundary
        for exon in samline.transcript.exons[1:]:
            if exon.start - exon_boundary <= samline.start and exon.start + exon_boundary >= samline.start:
                return samline.start - exon.start + exon.rstart
        return False
    def bp_position(self):
        if self.samline1.num < self.samline2.num:
            #gene1 -> gene2
            front_samline = self.samline1
            behind_samline = self.samline2
            self.anti = False
        else:
            #gene2 -> gene1
            front_samline = self.samline2
            behind_samline = self.samline1
            self.anti = True
        self.fgene = front_samline.transcript.gene
        self.bgene = behind_samline.transcript.gene
        self.fsam = front_samline
        self.bsam = behind_samline
        if front_samline.flag == 0 or front_samline.flag == 2048:
            if behind_samline.flag == 0 or behind_samline.flag == 2048:
                #++
                self.bp1 = self.front_bp_pos(front_samline)
                self.bp2 = self.behind_bp_pos(behind_samline)
                self.flag = 0
                pass
            else:
                #+-
                self.bp1 = self.front_bp_pos(front_samline)
                self.bp2 = self.front_bp_pos(behind_samline)
                self.flag = 1
                pass
        else:
            if behind_samline.flag == 0 or behind_samline.flag == 2048:
                #-+
                self.bp1 = self.behind_bp_pos(front_samline)
                self.bp2 = self.behind_bp_pos(behind_samline)
                self.flag = 2
                pass
            else:
                #--
                self.bp1 = self.behind_bp_pos(front_samline)
                self.bp2 = self.front_bp_pos(behind_samline)
                self.flag = 3
                pass
        if self.samline1.num > self.samline2.num:
            if self.flag == 0 or self.flag == 3:
                self.flag = 3 - self.flag
        if self.bp1 == False or self.bp2 == False:
            self.flag = 4
class Tree_node():
    def __init__(self, num, height, leaves):
        self.num = num
        self.height = height
        self.leaves = leaves
        self.father = None
        self.leftson = None
        self.rightson = None
        self.calculate_address()
        pass
    def add_group(self, group, group_flag):
        self.group = group
        self.gf = group_flag
        pass
    def calculate_address(self):
        address1 = 0
        address2 = 0
        for leaf in self.leaves:
            address1 += leaf.address[0]
            address2 += leaf.address[1]
        address1 = int(address1/len(self.leaves))
        address2 = int(address2/len(self.leaves))
        self.address = (address1, address2)
    def add_father(self, father):
        self.father = father
    def add_sons(self, son1, son2):
        self.leftson = son1
        self.rightson = son2
class Cluster_tree():
    def __init__(self, leaves, gene1, gene2, flag):
        self.leaflist = leaves
        self.flag = flag
        address1list1 = []
        address1list2 = []
        self.root = None
        self.nodes = {0 : []}
        for leaf in self.leaflist:
            if leaf.fgene == gene1:
                address1list1.append(leaf.bp1)
                address1list2.append(leaf.bp2)
            else:
                address1list1.append(leaf.bp2)
                address1list2.append(leaf.bp1)
        mean1 = np.mean(address1list1)
        mean2 = np.mean(address1list2)
        for i in range(len(self.leaflist)):
            leaf = self.leaflist[i]
            leaf.num = i
            leaf.add_address((address1list1[i] - mean1, address1list2[i] - mean2))
        num = 0
        for leaf in self.leaflist:
            node = Tree_node(num, 0, [leaf,])
            num += 1
            self.nodes[0].append(node)
        self.num = num - 1
        self.mean1 = int(mean1)
        self.mean2 = int(mean2)
        pass
    def clustering(self):
        n = 0
        while len(self.nodes[n]) > 1:
            self.nodes[n + 1] = self.find_min_distance(self.nodes[n])
            n += 1
        self.root = self.nodes[n][0]
        pass
    def find_min_distance(self, nodes):
        n = len(nodes)
        distance = self.calculate_distance(nodes[0].address, nodes[1].address)
        lnode = nodes[0]
        rnode = nodes[1]
        for i in range(n - 1):
            for j in range(i + 1, n):
                if distance > self.calculate_distance(nodes[i].address, nodes[j].address):
                    distance = self.calculate_distance(nodes[i].address, nodes[j].address)
                    lnode = nodes[i]
                    rnode = nodes[j]
        self.num += 1
        leaves = lnode.leaves + rnode.leaves
        fnode = Tree_node(self.num, distance, leaves)
        fnode.add_sons(lnode, rnode)
        lnode.add_father(fnode)
        rnode.add_father(fnode)
        newnodes = [fnode,]
        for node in nodes:
            if node != rnode and node != lnode:
                newnodes.append(node)
        return newnodes
    def calculate_distance(self, a, b):
        distance = abs(a[0] - b[0]) + abs(a[1] - b[1])
        return distance
    def print_tree(self, file):
        with open(file, 'w') as result:
            for i in range(len(self.nodes.keys())):
                n = len(self.nodes.keys()) - i - 1
                line0 = 'layer' + str(n) + ':\n'
                line1 = ''
                for node in self.nodes[n]:
                    line1 += 'node num: ' + str(node.num) + '\tdistance between sons: ' + str(node.height) + '\n' + 'leaf num: '
                    for leaf in node.leaves:
                        line1 += str(leaf.num) + ' '
                    line1 += '\naddress: ' + str(node.address[0] + self.mean1) + ', ' + str(node.address[1] + self.mean2) + '\n'
                result.write(line0)
                result.write(line1 + '\n')
        pass
    def classify(self, k):
        global sup_read_num
        for i in range(len(self.nodes.keys())):
            n = len(self.nodes.keys()) - i - 1
            if_classify = True
            for node in self.nodes[n]:
                if node.height > k:
                    if_classify = False
            if if_classify == True :
                min = sup_read_num
                nodelist = []
                for node in self.nodes[n]:
                    if len(node.leaves) >= min:
                        nodelist.append(node)
                return nodelist
'''
fuction: Antisense
input: string
output: antistring
'''
def Antisense(string):
    antistring = ''
    for i in string:
        if i == 'A':
            antistring += 'T'
        elif i == 'T':
            antistring += 'A'
        elif i == 'C':
            antistring += 'G'
        elif i == 'G':
            antistring += 'C'
    antistring = antistring[::-1]
    return antistring
'''
fuction: subprocess_popen
'''
def subprocess_popen(args, stdin=None, stdout=PIPE, stderr=sys.stderr, bufsize=8388608):
    return Popen(args, stdin=stdin, stdout=stdout, stderr=stderr, bufsize=bufsize, universal_newlines=True)
'''
step 1.1
fuction: read_annotation()
input: annotation
output: NULL (global index)
Read rebuilded annotation to make index for samlines to gene.
'''
def read_annotation(annotation):
    global index
    m = 0
    n = 0
    chr_to_genes = {}
    gtf = open(annotation)
    line = gtf.readline()
    line = gtf.readline()
    while line:
        if line.split('\t')[1] == 'gene':
            m += 1
            id = line.split('\t')[5]
            name = line.split('\t')[6]
            refstart = int(line.split('\t')[2].split(':')[0])
            refend = int(line.split('\t')[2].split(':')[1])
            chr = line.split('\t')[0]
            flag = line.split('\t')[4]
            gene_type = line.split('\t')[7].split('\n')[0]
            gene = Gene(id, name, refstart, refend, chr, flag, gene_type)
            if chr in chr_to_genes:
                chr_to_genes[chr].append(gene)
            else:
                chr_to_genes[chr] = [gene,]
            pass
        elif line.split('\t')[1] ==  'transcript':
            n += 1
            id = line.split('\t')[5]
            name = line.split('\t')[6]
            refstart = int(line.split('\t')[2].split(':')[0])
            refend = int(line.split('\t')[2].split(':')[1])
            start = int(line.split('\t')[3].split(':')[0])
            end = int(line.split('\t')[3].split(':')[1])
            chr = line.split('\t')[0]
            flag = line.split('\t')[4]
            trans_type = line.split('\t')[7].split('\n')[0]
            transcript = Transcript(id, name, refstart, refend, start, end, chr, flag, trans_type, gene)
            gene.add_transcript(transcript)
            pass
        elif line.split('\t')[1] == 'exon':
            id = line.split('\t')[5]
            refstart = int(line.split('\t')[2].split(':')[0])
            refend = int(line.split('\t')[2].split(':')[1])
            start = int(line.split('\t')[3].split(':')[0])
            end = int(line.split('\t')[3].split(':')[1])
            exon = Exon(id, refstart, refend, start, end, transcript)
            transcript.add_exon(exon)
            pass
        line = gtf.readline()
    for chr, genes in chr_to_genes.items():
        tree = IntervalTree()
        for gene in genes:
            for transcript in gene.transcripts:
                start = transcript.start
                end = transcript.end
                tree[start:end] = transcript
        index[chr] = tree
'''
step 1.2
fuction: read_sam_file()
input: thread, ref_index, readfile
output: NULL (global reads)
Use minimap2 to align reads to rebuild transcripts and then index the result of alignment.
'''
def read_sam_file(thread, ref_index, readfile,  output_sam, if_output, type_alignment):
    global reads, overlap_present
    minimap2_view_process = subprocess_popen(shlex.split('minimap2 -ax map-{} -t {} {} {} --secondary=no'.format(type_alignment, str(thread), ref_index, readfile)))
    old_id = ''
    n = 0
    with open(output_sam, 'w') as samfile:
        for line in minimap2_view_process.stdout:
            if if_output == True:
                samfile.write(line)
            if 'SA:' in line and line[0] != '@':
                read_id = line.split('\t')[0]
                read_flag = int(line.split('\t')[1])
                if read_flag != 4 and read_flag != 256 and read_flag != 272:
                    read_chr = line.split('\t')[2]
                    read_start = int(line.split('\t')[3])
                    read_cigar = line.split('\t')[5]
                    read_fasta = line.split('\t')[9]
                    if read_id != old_id:
                        old_id = read_id
                        n += 1
                        if n > 1 and len(read.samlines) > 1:
                            if read_to_transcript(read, overlap_present) == True:
                                reads.append(read)
                        read = Read(read_id, read_start, read_cigar, read_flag, read_chr, read_fasta)
                    else:
                        read.add_samline(read_id, read_start, read_cigar, read_flag, read_chr, read_fasta)
    if len(read.samlines) > 1:
        if read_to_transcript(read, overlap_present) == True:
            reads.append(read)
    minimap2_view_process.stdout.close()
    minimap2_view_process.wait()

def read_sam_file2(output_sam):
    global reads, overlap_present
    old_id = ''
    n = 0
    samfile = open(output_sam)
    line = samfile.readline()
    while line:
        if 'SA:' in line and line[0] != '@':
            read_id = line.split('\t')[0]
            read_flag = int(line.split('\t')[1])
            if read_flag != 4 and read_flag != 256 and read_flag != 272:
                read_chr = line.split('\t')[2]
                read_start = int(line.split('\t')[3])
                read_cigar = line.split('\t')[5]
                read_fasta = line.split('\t')[9]
                if read_id != old_id:
                    old_id = read_id
                    n += 1
                    if n > 1 and len(read.samlines) > 1: #filter3
                        if read_to_transcript(read, overlap_present) == True:
                            reads.append(read)
                    read = Read(read_id, read_start, read_cigar, read_flag, read_chr, read_fasta)
                else:
                    read.add_samline(read_id, read_start, read_cigar, read_flag, read_chr, read_fasta)
        line = samfile.readline()
    if len(read.samlines) > 1:
        if read_to_transcript(read, overlap_present) == True:
            reads.append(read)
    pass

def read_to_transcript(read, precent):
    global index
    for samline in read.samlines:
        Sort = sorted(index[samline.chr][samline.start:samline.end])
        for m in Sort:
            transcriptlength = m[1] - m[0] + 1
            cigarlength = samline.end - samline.start + 1
            if transcriptlength < cigarlength:
                limit = int(transcriptlength * precent)
            else:
                limit = int(cigarlength * precent)
            if samline.end - m[0] >= limit and m[1] - samline.start >= limit:
                samline.add_transcript(m[2], m[0], m[1])
    
    if read.transcripts_and_genes() == True:
        read.num_samline()
        if len(read.genes) > 1:
            return True
    return False

'''
step 1.3
fuction: gene_groups()
input: read_in_same_chr
output: (gene_groups_with_one_read, gene_groups_with_multiple_read)
Find gene pairs with the number of supporting reads larger than one. 
'''               
def gene_groups():
    global reads
    time1 = time.time()
    m = 0
    print('Sparse matrix multiplication-based candidate gene pairing')
    gene_groups = []
    gene_to_vector = {} #{key = gene : value = [read, ...]}
    gene_to_reads = {}
    num_of_reads = len(reads)
    vector_list = []
    for i in range(num_of_reads):
        read = reads[i]
        for gene in list(read.genes.keys()):
            if gene in gene_to_vector:
                gene_to_vector[gene][i] = 1
                gene_to_reads[gene].append(read)
            else:
                gene_to_vector[gene] = np.zeros(num_of_reads, dtype= np.uint8)
                gene_to_vector[gene][i] = 1
                gene_to_reads[gene] = [read,]
    num_of_genes = len(gene_to_vector.keys())
    for gene, vector in gene_to_vector.items():
        vector_list.append(vector)
    gene_Matrix = np.vstack(vector_list)
    csr_gene_Matrix = csr_matrix(gene_Matrix)
    csr_gene_Matrix_T = csr_matrix(gene_Matrix.T)
    groups = csr_gene_Matrix.dot(csr_gene_Matrix_T)
    count = groups.indptr
    index = groups.indices
    data = groups.data
    for i in range(num_of_genes - 1):
        index0 = index[count[i] : count[i + 1]]
        data0 = data[count[i] : count[i + 1]]
        if index0.size > 0:
            for j in range(count[i + 1] - count[i]):
                if data0[j] > 1 and index0[j] > i:
                    gene1 = list(gene_to_vector.keys())[i]
                    gene2 = list(gene_to_vector.keys())[index0[j]]
                    read_list = []
                    add_vector = gene_to_vector[gene1] + gene_to_vector[gene2]
                    index_read = np.where(add_vector == 2)
                    for k in index_read[0]:
                        read_list.append(reads[k])
                    genegroup = Genes_group(gene1= gene1, gene2= gene2, reads= read_list)
                    gene_groups.append(genegroup)
                    m += 1
    time2 = time.time()
    print('used ', time2 - time1, 's')
    return gene_groups
'''
step 1.output_result
fuction: output_result()
input: genegroups, output_file_dir
output: NULL
Output the gene groups to dir.
'''
def output_result(genegroups, output_file_dir):
    time1 = time.time()
    print('-' * 50)
    print('Outputing result')
    if not os.path.exists(output_file_dir):
        os.makedirs(output_file_dir)
    for group in genegroups:
        file_name = group.gene1.name + '::' + group.gene2.name + '.txt'
        with open(output_file_dir + file_name, 'w') as outputfile:
            outputfile.write('@' + group.gene1.id + '::' + group.gene2.id + '\treads num: ' +str(len(group.reads)) + '\n')
            for read in group.reads:
                outputfile.write('>' + read.id + '\n')
                line0 = 'chr\tgene id\ttranscript id\ttstart:end\tstart:end\tposition in read\n'
                length1 = 0
                length2 = 0
                for samline in read.samlines:
                    if samline.transcript.gene.id == group.gene1.id:
                        if samline.length > length1:
                            length1 = samline.length
                            line1 = samline.chr + '\t' + samline.transcript.gene.id + '\t' + samline.transcript.id + '\t' + \
                                str(samline.tstart) + ':' + str(samline.tend) + '\t' + \
                                str(samline.start) + ':' + str(samline.end) + '\t' + str(samline.num) + '\t' + samline.sequence + '\n'
                    elif samline.transcript.gene.id == group.gene2.id:
                        if samline.length > length2:
                            length2 = samline.length
                            line2 = samline.chr + '\t' + samline.transcript.gene.id + '\t' + samline.transcript.id + '\t' + \
                                str(samline.tstart) + ':' + str(samline.tend) + '\t' + \
                                str(samline.start) + ':' + str(samline.end) + '\t' + str(samline.num) + '\t' + samline.sequence + '\n'
                outputfile.write(line0)
                outputfile.write(line1)
                outputfile.write(line2)
    time2 = time.time()
    print('used ', time2 - time1, 's')
'''
step 1.4
fuction: clustering
input: genegroups
output: NULL
Cluster the reads' breakpoints in one gene group.
'''
def clustering(genegroups, tree_dir, out_put, result_out, if_output):
    time1 = time.time()
    print('-' * 50)
    print('Clustering the breakpoint and POA')
    global clustering_num
    nodes0 = []
    n_list = []
    result_file = ''
    POA_file = ''
    with open(out_put, 'w') as output:
        for group in genegroups:
            group.clustering()
            if group.flag == True:
                for flag, tree in group.trees.items():
                    if if_output == True:
                        file_name = group.gene1.name + '::' + group.gene2.name + '_' + str(tree.flag) + '.txt'
                        if not os.path.exists(tree_dir):
                            os.makedirs(tree_dir)
                        tree.print_tree(tree_dir + file_name)
                    nodelist = tree.classify(clustering_num)
                    n = 0
                    for node in nodelist:
                        node.add_group(group, flag)
                        nodes0.append(node)
                        read_list = []
                        read_list_1 = []
                        read_list_2 = []
                        
                        leaf = node.leaves[0]
                        if leaf.anti == True:
                            min_f = len(leaf.bsam.sequence)
                            min_b = len(leaf.fsam.sequence)
                        else:
                            min_b = len(leaf.bsam.sequence)
                            min_f = len(leaf.fsam.sequence)
                        
                        for leaf in node.leaves:
                            if leaf.anti == True:
                                if len(leaf.bsam.sequence) < min_f:
                                    min_f = len(leaf.bsam.sequence)
                                if len(leaf.fsam.sequence) < min_b:
                                    min_b = len(leaf.fsam.sequence)
                                read_list_1.append(Antisense(leaf.bsam.sequence))
                                read_list_2.append(Antisense(leaf.fsam.sequence))
                            else:
                                if len(leaf.fsam.sequence) < min_f:
                                    min_f = len(leaf.fsam.sequence)
                                if len(leaf.bsam.sequence) < min_b:
                                    min_b = len(leaf.bsam.sequence)
                                read_list_1.append(leaf.fsam.sequence)
                                read_list_2.append(leaf.bsam.sequence)
                        min_f_seq = max(min_abpoa, min_f)
                        min_b_seq = max(min_abpoa, min_b)
                        for i in range(len(read_list_1)):
                            seq1 = read_list_1[i]
                            seq2 = read_list_2[i]
                            if min_f_seq < len(seq1):
                                seq1 = seq1[-min_f_seq:]
                            if min_b_seq < len(seq2):
                                seq2 = seq2[:min_b_seq]
                            read_list.append(seq1 + seq2)
                        a = pa.msa_aligner()
                        res=a.msa(read_list, out_cons=True, out_msa=True)
                        for seq in res.cons_seq:
                            line1 = seq + '\n'
                        n += 1
                        line0 = '>' + group.gene1.id + '_' + group.gene2.id + '_' + str(n) + '_' + str(len(node.leaves)) + '\n'
                        output.write(line0)
                        output.write(line1)
                        POA_file += line0 + line1
                        n_list.append(n)
    time2 = time.time()
    print('find ', str(len(nodes0)), 'fusion genes groups')
    print('used ', time2 - time1, 's')
    time1 = time.time()
    print('-' * 50)
    print('Output result')
    groups = {}
    with open(result_out, 'w') as file:
        if if_output == True:
            file.write('gene1 name\tgene2 name\tgene1 id\tgene2 id\tchr1\tchr2\tgene1 pos\tgene2 pos\tbreakpoint1\tbreakpoint2\tnum\tcode\treads\n')
        i = 0
        for node in nodes0:
            code = str(n_list[i])
            i += 1
            gene1 = node.group.gene1.id
            gene2 = node.group.gene2.id
            n1 = node.group.gene1.name
            n2 = node.group.gene2.name
            pos1 = str(node.group.gene1.rstart) + ':' + str(node.group.gene1.rend)
            pos2 = str(node.group.gene2.rstart) + ':' + str(node.group.gene2.rend)
            chr1 = node.group.gene1.chr + ':' + node.group.gene1.flag
            chr2 = node.group.gene2.chr + ':' + node.group.gene2.flag
            read_num = len(node.leaves)
            break1 = int(node.address[0] + node.group.trees[node.gf].mean1)
            break2 = int(node.address[1] + node.group.trees[node.gf].mean2)

            key = '>' + group.gene1.id + '_' + group.gene2.id + '_' + str(n) + '_' + str(len(node.leaves))
            reads_ls = []
            reads_name = ''
            for leaf in node.leaves:
                reads_ls.append(leaf.read)
                reads_name += leaf.read.id+ ';'

            line = n1 + '\t' + n2 + '\t' + \
                gene1 + '\t' + gene2 + '\t' + \
                chr1 + '\t' + chr2 + '\t' + \
                pos1 + '\t' + pos2 + '\t' + \
                str(break1) + '\t' + str(break2) + '\t' + \
                str(read_num) + '\t' + code + '\t' + reads_name +'\n'
            if if_output == True:
                file.write(line)
            result_file += line
            if key not in groups:
                groups[key] = reads_ls
    time2 = time.time()
    print('used ', time2 - time1, 's')
    print('-' * 50)
    return result_file, POA_file, groups
'''
fuction: usage_and_exit
'''
def usage_and_exit():
    sys.stderr.write('Usage: genefusion [indexdir]')
    sys.stderr.write('\n')
    exit(0)
'''
fuction input_files
'''
def input_files():
    if len(sys.argv) < 3:
        usage_and_exit()
    else:
        annotation = sys.argv[1] + 'rebuild_annotation.txt'
        ref_index = sys.argv[1] + 'trans_index.mmi'
        readfile = sys.argv[2]
        if len(sys.argv) > 3:
            thread = int(sys.argv[3])
        else:
            thread = 2
    return (annotation, ref_index, readfile, thread)

'''
fuction: control_thread
input: annotation, ref_index, readfile, thread, if_output, link_tuple
output: NULL
Make sure the thread number is posible.
'''
def control_thread(annotation, ref_index, readfile, thread, if_output, link_tuple, cnums, if_countine):
    global index
    global reads
    global min_length, exon_boundary, overlap_present, clustering_num, sup_read_num
    output_sam, output_file_dir, tree_dir, out_put, result_out = link_tuple
    min_length, exon_boundary, overlap_present, clustering_num , type_alignment, sup_read_num= cnums
    cpu = os.cpu_count()
    if cpu < thread:
        thread = cpu
    time1 = time.time()
    print('-' * 50)
    print('Reading index file')
    read_annotation(annotation)
    time2 = time.time()
    print('used ', time2 - time1, 's')
    print('-' * 50)
    time1 = time.time()
    print('Transcriptome-based alignment')
    print('-' * 50)
    if if_countine == False:
        read_sam_file(thread, ref_index, readfile, output_sam, if_output, type_alignment)
    else:
        read_sam_file2(output_sam)
    time2 = time.time()
    print('found ', len(reads), ' reads with genefusion')
    print('used ', time2 - time1, 's')
    print('-' * 50)
    genegroups = gene_groups()
    if if_output == True:
        output_result(genegroups, output_file_dir)
    result_file, POA_file, groups = clustering(genegroups, tree_dir, out_put, result_out, if_output)
    return result_file, POA_file, groups
'''
fuction: main
'''
def main(readfile, index_dir, middlefile, thread, if_output, cnums, if_countine):
    annotation = index_dir + 'rebuild_annotation.txt'
    ref_index = index_dir + 'trans_index.mmi'
    output_sam = middlefile + 'step1.DNA_like_alignment.sam'
    output_file_dir = middlefile + 'step1.gene_groups/'
    tree_dir = middlefile + 'step1.clustering_tree/'
    out_put = middlefile + 'step1.POA_result.fasta'
    result_out = middlefile + 'step1.genefusion.csv'
    link_tuple = (output_sam, output_file_dir, tree_dir, out_put, result_out)
    time1 = time.time()
    result_file, POA_file, groups = control_thread(annotation, ref_index, readfile, thread, if_output, link_tuple, cnums, if_countine)
    time2 = time.time()
    timeuse = time2 - time1
    print('Transcriptome-based detection totally used ' + str(timeuse) + 's')
    print('-' * 50)
    return result_file, POA_file, groups
'''
main fuction
'''
if __name__ == "__main__":
    annotation, ref_index, readfile, thread = input_files()
    link_tuple = (output_sam, output_file_dir, tree_dir, out_put, result_out)
    print('-' * 50)
    print('Start detecting!!')
    time1 = time.time()
    cnums = (min_length, exon_boundary, overlap_present, clustering_num, type_alignment)
    control_thread(annotation, ref_index, readfile, thread, True, link_tuple, cnums)
    time2 = time.time()
    timeuse = time2 - time1
    print('Totally using ' + str(timeuse) + 's')
    print('-' * 50)


import time
import os, sys

rebuild_transcript_file = 'rebuild_transcript.fa'
rebuild_index_file = 'rebuild_annotation.txt'

'''
class: Gene
'''
class Gene():
    def __init__(self, id, name, start, end, chr, flag, gene_type):
        self.id = id
        self.name = name
        self.start = start
        self.end = end
        self.chr = chr
        self.flag = flag
        self.type = gene_type
        self.transcripts = []
        pass
    def add_transcript(self, transcript):
        self.transcripts.append(transcript)
        pass
'''
class: Transcript
'''
class Transcript():
    def __init__(self, id, name, start, end, chr, flag, trans_type):
        self.id = id
        self.name = name
        self.start = start
        self.end = end
        self.chr = chr
        self.flag = flag
        self.trans_type = trans_type
        self.exons = []
        pass
    def add_exon(self, exon):
        self.exons.append(exon)
        pass
    def add_pos(self, start, end):
        self.restart = start
        self.reend = end
'''
class: Exon
'''
class Exon():
    def __init__(self, id, start, end):
        self.id = id
        self.start = start
        self.end = end
        pass
    def add_sequence(self, sequence):
        self.sequence = sequence
        pass
    def add_pos(self, start, end):
        self.restart = start
        self.reend = end
'''
fuction: Read_Annotation
input: gtffile
output: chr_to_genes #{chr : [gene1, gene2, ...]}
'''
def Read_Annotation(gtffile):
    print('-' * 50)
    print('Reading annotation file')
    time1 = time.time() 

    m = 0
    n = 0
    chr_to_genes = {}
    gtf = open(gtffile)
    line = gtf.readline()
    while line:
        if '##' not in line:
            if line.split('\t')[2] == 'gene':
                m += 1
                id = line.split('\t')[8].split(';')[0].split('"')[1]
                name = line.split('\t')[8].split(';')[2].split('"')[1]
                start = int(line.split('\t')[3])
                end = int(line.split('\t')[4])
                chr = line.split('\t')[0]
                flag = line.split('\t')[6]
                gene_type = line.split('\t')[8].split(';')[1].split('"')[1]
                gene = Gene(id, name, start, end, chr, flag, gene_type)
                if chr in chr_to_genes:
                    chr_to_genes[chr].append(gene)
                else:
                    chr_to_genes[chr] = [gene,]
                pass
            elif line.split('\t')[2] ==  'transcript':
                n += 1
                id = line.split('\t')[8].split(';')[1].split('"')[1]
                name = line.split('\t')[8].split(';')[5].split('"')[1]
                start = int(line.split('\t')[3])
                end = int(line.split('\t')[4])
                chr = line.split('\t')[0]
                flag = line.split('\t')[6]
                trans_type = line.split('\t')[8].split(';')[4].split('"')[1]
                transcript = Transcript(id, name,start, end, chr, flag, trans_type)
                gene.add_transcript(transcript)
                pass
            elif line.split('\t')[2] == 'exon':
                id = line.split('\t')[8].split(';')[6].split(' ')[2]
                start = int(line.split('\t')[3])
                end = int(line.split('\t')[4])
                exon = Exon(id, start, end)
                transcript.add_exon(exon)
                pass
        line = gtf.readline()
    
    time2 = time.time()
    timeuse = time2 - time1
    print('Using ' + str(timeuse) + 's')
    print('Read ', m, ' genes')
    print('Read ', n, ' transcriprts')
    print('-' * 50)
    return chr_to_genes
'''
fuction: Filtering
input: chr_to_genes #{chr : [gene1, gene2, ...]}
output: NULL
'''
def Filtering(chr_to_genes):
    print('-' * 50)
    print('Filtering the genes')
    time1 = time.time() 
    new_chr_to_genes = {}
    m = 0
    n = 0
    for chr, genes in chr_to_genes.items():
        if chr != 'chrM' and chr != 'chrY':
            new_chr_to_genes[chr] = []
            for gene in genes:
                if gene.type == 'protein_coding':
                    new_chr_to_genes[chr].append(gene)
                    m += 1
                    for transcript in gene.transcripts:
                        n += 1
                        if gene.flag == '-':
                            transcript.exons = transcript.exons[::-1]
                        '''for exon in transcript.exons:
                            pass'''
    
    time2 = time.time()
    timeuse = time2 - time1
    print('Using ' + str(timeuse) + 's')
    print('Reserved ', m, ' genes')
    print('Reserved ', n, ' transcriprts')
    print('-' * 50)
    return new_chr_to_genes
'''
fuction: Read_Reference
input: reference, chr_to_genes
output: NULL
'''
def Read_Reference(reference, chr_to_genes):
    print('-' * 50)
    print('Read reference file')
    time1 = time.time() 

    chrs = chr_to_genes.keys()
    ref = open(reference)
    line = ref.readline()
    while line:
        if '>' in line:
            chr = line.split('>')[1].split(' ')[0]
            if chr in chrs:
                sequence = ''
                line = ref.readline()
                while '>' not in line:
                    sequence += line.split('\n')[0]
                    line = ref.readline()
                for gene in chr_to_genes[chr]:
                    for transcript in gene.transcripts:
                        for exon in transcript.exons:
                            exon.add_sequence(sequence[exon.start - 1: exon.end])
            else:
                line = ref.readline()
        else:
            line = ref.readline()
    
    time2 = time.time()
    timeuse = time2 - time1
    print('Using ' + str(timeuse) + 's')
    print('-' * 50)
'''
fuction: Build_transcript_file
input: chr_to_genes, rebuild_transcript_file, rebuild_index_file, middlefile
output: NULL
'''
def Build_transcript_file(chr_to_genes, rebuild_transcript_file, rebuild_index_file, middlefile):
    print('-' * 50)
    print('Building transcript file and index')
    time1 = time.time() 

    if not os.path.exists(middlefile):
        os.makedirs(middlefile)
    with open(middlefile + rebuild_transcript_file, 'w') as rtf:
        for chr, genes in chr_to_genes.items():
            rtf.write('>' + chr + '\n')
            refline = ''
            length = 0
            for gene in genes:
                for transcript in gene.transcripts:
                    tstart = length + 1
                    for exon in transcript.exons:
                        start = length + 1
                        refline += exon.sequence
                        length += exon.end - exon.start + 1
                        end = length
                        exon.add_pos(start, end)
                    tend = length
                    transcript.add_pos(tstart, tend)
            one_line = ''
            for i in refline:
                one_line += i
                if len(one_line) == 100:
                    rtf.write(one_line + '\n')
                    one_line = ''
            if one_line != '':
                rtf.write(one_line + '\n')
    with open(middlefile + rebuild_index_file, 'w') as rif:
        line0 = 'chr\tclass\trefpos\trebuildpos\tflag\tid\tname\ttype\n'
        rif.write(line0)
        for chr, genes in chr_to_genes.items():
            length = 0
            for gene in genes:
                line = gene.chr + '\tgene\t' + \
                    str(gene.start) + ':' + str(gene.end) + \
                    '\t-\t' + \
                    gene.flag + '\t' + gene.id + '\t' + gene.name + '\t' + \
                    gene.type + '\n'
                rif.write(line)
                for transcript in gene.transcripts:
                    line = transcript.chr + '\ttranscript\t' + \
                        str(transcript.start) + ':' + str(transcript.end) + '\t' + \
                        str(transcript.restart) + ':' + str(transcript.reend) + '\t' + \
                        transcript.flag + '\t' + transcript.id + '\t' + transcript.name + '\t' + \
                        transcript.trans_type+'\n'
                    rif.write(line)
                    for exon in transcript.exons:
                        line = transcript.chr + '\texon\t' + \
                            str(exon.start) + ':' + str(exon.end) + '\t' + \
                            str(exon.restart) + ':' + str(exon.reend) + '\t' + \
                            transcript.flag + '\t' + exon.id +'\n'
                        rif.write(line)
    
    time2 = time.time()
    timeuse = time2 - time1
    print('Using ' + str(timeuse) + 's')
    print('-' * 50)
'''
fuction: usage_and_exit
'''
def usage_and_exit():
    sys.stderr.write('Usage: index [annotationfile] [referencefile] [indexdir]')
    sys.stderr.write('\n')
    exit(0)
'''
fuction: input_files
'''
def input_files():
    if len(sys.argv) < 4:
        usage_and_exit()
    else:
        gtffile = sys.argv[1]
        reference = sys.argv[2]
        middlefile = sys.argv[3]
    return (gtffile, reference, middlefile)
'''
fuction: main
'''
def main(gtffile, reference, middlefile):
    chr_to_genes = Read_Annotation(gtffile)
    chr_to_genes = Filtering(chr_to_genes)
    Read_Reference(reference, chr_to_genes)
    Build_transcript_file(chr_to_genes, rebuild_transcript_file, rebuild_index_file, middlefile)
'''
main fuction
'''
if __name__ == "__main__":
    gtffile, reference, middlefile = input_files()
    chr_to_genes = Read_Annotation(gtffile)
    chr_to_genes = Filtering(chr_to_genes)
    Read_Reference(reference, chr_to_genes)
    Build_transcript_file(chr_to_genes, rebuild_transcript_file, rebuild_index_file, middlefile)


from setuptools import setup, find_packages

setup(
    name='GFHunter',
    version='1.0.1',
    description=(
    	'GFHunter is a long read sequence transcriptome alignment-based fusion genes detection tool.'
    	 ),
    long_description=open('README.md').read(),
    author='Zhenhao Lu',
    author_email='luzhenhao.hit@gmail.com',
    maintainer='Zhenhao Lu',
    maintainer_email='luzhenhao.hit@gmail.com',
    license='MIT License',
    url='https://github.com/luzhenhao-HIT/GFHunter',
    packages=find_packages(),
    entry_points={
    'console_scripts': [
        'GFHunter = GFHunter.main:Parser_set',
        ]
    },
    classifiers=[
        'Operating System :: OS Independent',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3'
    ],
    python_requires='>=3',
    install_requires=[
		'numpy',
        'pysam',
		'scipy',
		'intervaltree',
		'pyabpoa'
    ]
)



import pysam
import os
import time

class Cell():
    def __init__(self, CB):
        self.CB = CB
        self.reads = []
        self.fusions = []
        pass
    def add_reads(self, read):
        self.reads.append(read)
    def add_fusions(self, fusion):
        if fusion not in self.fusions:
            self.fusions.append(fusion)

def read_bam(outfile, bam, output):
    time1 = time.time()
    f_name, e = os.path.splitext(os.path.basename(bam))
    indexfile = outfile + f_name + '.CBindex'
    fasta = output + '.fasta'
    bamfile = pysam.AlignmentFile(bam, 'rb')
    anti_index = {}
    if not os.path.exists(indexfile):
        with open(fasta, 'w') as fasta_out:
            for read in bamfile:
                CB = read.get_tag('CB')
                UMI = read.get_tag('UB')
                if CB in anti_index:
                    anti_index[CB].append((read.query_name, UMI))
                else:
                    anti_index[CB] = [(read.query_name, UMI),]
                if read.flag in [0, 16]:
                    fasta_out.write(f">{read.query_name}\n{read.query_sequence}\n")
                

        with open(indexfile, 'w') as f:
            f.write('CB\tread\n')
            for CB, reads in anti_index.items():
                for read in reads:
                    name, UMI = read
                    line = CB + '\t' + name + '\t' + UMI + '\n'
                    f.write(line)
    else:
        print('Index exists, jump to the next step')
    time2 = time.time()
    timeuse = time2 - time1
    print('Indexing time used: ', timeuse)

def create_key(a, b):
    ls = [a, b]
    ls.sort()
    key = ls[0] + ':' + ls[1]
    return key

def read_index(file):
    index = {}
    cells = {}
    f = open(file)
    line = f.readline()
    while line:
        CB = line.split('\t')[0]
        read = line.split('\t')[1].split('\n')[0]
        if CB in cells:
            cells[CB].add_reads(read)
        else:
            cell = Cell(CB)
            cells[CB] = cell
        index[read] = cells[CB]
        line = f.readline()
    return index

def read_fusions(file, index, output, lc):
    fusion_cells = []
    fusion_line = {}
    result = open(file)
    line = result.readline()
    line = result.readline()
    while line:
        g1 = line.split('\t')[0]
        g2 = line.split('\t')[1]
        id1 = line.split('\t')[2]
        id2 = line.split('\t')[3]
        bps = line.split('\t')[4]
        gp1 = line.split('\t')[5]
        gp2 = line.split('\t')[6]
        rnum = line.split('\t')[7]
        flag = line.split('\t')[8]
        reads_l = line.split('\t')[9]
        key = create_key(g1, g2)
        reads = line.split('\t')[9].split('\n')[0].split(';')[:-1]
        cells = []
        for read in reads:
            cell = index[read]
            cells.append(cell.CB)
            if flag == 'flag = 5: Reliable Fusion':
                cell.add_fusions(key)
                if cell not in fusion_cells:
                    fusion_cells.append(cell)
        cs = list(set(cells))
        cell_l = ''
        for c in cs:
            cell_l += c + ';'
        new_line = g1 + '\t' + g2 + '\t' + id1 + '\t' + id2 + '\t' + bps + '\t' + gp1 + '\t' + gp2 + '\t' + str(len(cs)) + '\t' + rnum + '\t' + flag + '\t' + cell_l + '\t' + reads_l
        if key not in fusion_line:
            fusion_line[key] = [new_line,]
        else:
            fusion_line[key].append(new_line)
        line = result.readline()

    with open(output + '.sc.tsv', 'w') as outfile:
        outfile.write('gene1 name\tgene2 name\tgene1 id\tgene2 id\tbreak points\tgene1 position\tgene2 position\t#cells\t#reads\trank class\tcells\treads\n')
        for fusion, lines in fusion_line.items():
            sum_c_num = 0
            for line in lines:
                c_num = int(line.split('\t')[7])
                sum_c_num += c_num
            if sum_c_num >= lc:
                for line in lines:
                    outfile.write(line)
    return fusion_cells

def output_result(fusion_cells, file, lc):
    index = {}
    with open(file + '.r1.tsv', 'w') as rs:
        rs.write('cell\tfusions\n')
        for cell in fusion_cells:
            l = ''
            for f in cell.fusions:
                l += f + ';'
                if f in index:
                    index[f].append(cell.CB)
                else:
                    index[f] = [cell.CB,]
            line = cell.CB + '\t' + l + '\n'
            rs.write(line)
    sorted_index = dict(sorted(index.items(), key=lambda item: len(item[1]), reverse=True))
    with open(file + '.r2.tsv', 'w') as rs:
        rs.write('fusion\tcell\n')
        for f, cells in sorted_index.items():
            if len(cells) < lc:
                break
            for c in cells:
                line = f + '\t' + c + '\n'
                rs.write(line)
    with open(file + '.sum.tsv', 'w') as rs:
        rs.write('fusion\t#cell\tcells\n')
        for f, cells in sorted_index.items():
            if len(cells) < lc:
                break
            l = ''
            for c in cells:
                l += c + ';'
            line = f + '\t' + str(len(cells)) + '\t' + l + '\n'
            rs.write(line)

def handle(indexfile, bam, output, lc):
    f_name, e = os.path.splitext(os.path.basename(bam))
    indexf = indexfile + f_name + '.CBindex'
    result = output + '.tsv'
    index = read_index(indexf)
    fusion_cells = read_fusions(result, index, output, lc)
    output_result(fusion_cells, output, lc)